from __future__ import absolute_import
from libloader.com import load_com
from .base import Output, OutputError
import pywintypes


class WindowEyes(Output):
    """Speech output supporting the WindowEyes screen reader"""

    name = "Window-Eyes"

    def __init__(self, *args, **kwargs):
        super(WindowEyes, self).__init__(*args, **kwargs)
        try:
            self.object = load_com("gwspeak.speak")
        except (pywintypes.com_error, TypeError):
            raise OutputError

    def speak(self, text, interrupt=0):
        if interrupt:
            self.silence()
        self.object.SpeakString(text)

    def silence(self):
        self.object.Silence()

    def is_active(self):
        try:
            import win32gui
        except ImportError:
            return False
        try:
            return win32gui.FindWindow("GWMExternalControl", "External Control") != 0
        except:
            return False


output_class = WindowEyes



from __future__ import absolute_import
from libloader.com import load_com
import pywintypes

from .base import Output, OutputError


class Jaws(Output):
    """Supports the Jaws for Windows screen reader."""

    name = "jaws"

    def __init__(self, *args, **kwargs):
        super(Jaws, self).__init__(*args, **kwargs)
        try:
            self.object = load_com("FreedomSci.JawsApi", "jfwapi")
        except (pywintypes.com_error, TypeError):
            raise OutputError

    def braille(self, text, **options):
        # HACK: replace " with ', Jaws doesn't seem to understand escaping them with \
        text = text.replace('"', "'")
        self.object.RunFunction('BrailleString("%s")' % text)

    def speak(self, text, interrupt=False):
        self.object.SayString("      %s" % text, interrupt)

    def is_active(self):
        try:
            import win32gui
        except ImportError:
            return False
        try:
            return (
                self.object.SayString("", 0) == True
                or win32gui.FindWindow("JFWUI2", "JAWS") != 0
            )
        except:
            return False


output_class = Jaws



from platform import system as get_current_system
from .base import Output

class MacSpeech(Output):

    """Speech output supporting Apple MacOS NSSpeechSynthesizer."""

    name = "MacSpeech"
    priority = 101
    system_output = True

    def __init__(self, *args, **kwargs):
        from AppKit import NSSpeechSynthesizer
        self.synth = NSSpeechSynthesizer.alloc().init()

    def speak(self, text: str, interrupt: bool = False) -> bool:
        if interrupt:
            self.silence()
        return self.synth.startSpeakingString_(text)

    def silence(self):
        self.synth.stopSpeaking()

    def is_active(self):
        return get_current_system() == "Darwin"


output_class = MacSpeech



from __future__ import absolute_import
import ctypes
import os
import types
from platform_utils import paths


def load_library(libname, cdll=False):
    if paths.is_frozen():
        libfile = os.path.join(
            paths.embedded_data_path(), "accessible_output3", "lib", libname
        )
    else:
        libfile = os.path.join(paths.module_path(), "lib", libname)
    if not os.path.exists(libfile):
        _cxfreeze_libfile = os.path.join(
            paths.embedded_data_path(), "lib", "accessible_output3", "lib", libname
        )
        if os.path.exists(_cxfreeze_libfile):
            libfile = _cxfreeze_libfile
    if cdll:
        return ctypes.cdll[libfile]
    return ctypes.windll[libfile]


def get_output_classes():
    from . import outputs

    module_type = types.ModuleType
    classes = [
        m.output_class
        for m in outputs.__dict__.values()
        if isinstance(m, module_type) and hasattr(m, "output_class")
    ]
    return sorted(classes, key=lambda c: c.priority)


def find_datafiles():
    import platform
    from glob import glob
    import accessible_output3

    if platform.system() != "Windows":
        return []
    path = os.path.join(accessible_output3.__path__[0], "lib", "*.dll")
    results = glob(path)
    dest_dir = os.path.join("accessible_output3", "lib")
    return [(dest_dir, results)]



from __future__ import absolute_import
from .base import Output

class SpeechDispatcher(Output):
    """Supports Speech Dispatcher on Linux

    Note this requires python-speechd or python3-speechd to be installed.
    This can be done on Debian distros by using apt-get install python3-speechd.
    """

    name = "Linux Speech Dispatcher"
    _client = None
    _is_speaking = False

    def __init__(self):
        global speechd
        try:
            import speechd
        except:
            print("Cannot find Speech Dispatcher. Please install python-speechd or python3-speechd.")
        else:
            try:
                self._client = speechd.SSIPClient("")
            except Exception as e:
                print(e)

    def is_active(self):
        return self._client is not None

    def is_speaking(self):
        return self._is_speaking

    def _callback(self, callback_type):
        if callback_type == speechd.CallbackType.BEGIN:
            self._is_speaking = True
        elif callback_type == speechd.CallbackType.END:
            self._is_speaking = False
        elif callback_type == speechd.CallbackType.CANCEL:
            self._is_speaking = False

    def speak(self, text, interrupt=0):
        if interrupt:
            self.silence()
        self._client.speak(text, callback=self._callback,
                           event_types=(speechd.CallbackType.BEGIN,
                                        speechd.CallbackType.CANCEL,
                                        speechd.CallbackType.END))

    def silence(self):
        self._client.cancel()

    def close(self):
        # With speechd < 1.10.2, this method must be called
        # for the program to close correctly.
        self._client.close()


output_class = SpeechDispatcher



from __future__ import absolute_import
import ctypes
from .base import Output


class PCTalker(Output):
    """Supports the PC Talker screen reader."""

    name = "PC Talker"
    lib32 = "pctkusr.dll"
    lib64 = "pctkusr64.dll"
    cdll = True
    argtypes = {"PCTKPRead": (ctypes.c_char_p, ctypes.c_int, ctypes.c_int)}

    def speak(self, text, interrupt=False):
        if interrupt:
            self.silence()
        self.lib.PCTKPRead(text.encode("cp932", "replace"), 0, 1)

    def silence(self):
        self.lib.PCTKVReset()

    def is_active(self):
        return self.lib.PCTKStatus() != 0


output_class = PCTalker



from __future__ import absolute_import
from builtins import range
from libloader.com import load_com
from .base import Output

import logging

log = logging.getLogger(__name__)


class Sapi4(Output):
    """Supports the microsoft speech API version 4."""

    name = "sapi4"
    priority = 102

    def __init__(self):
        sapi4 = load_com("{EEE78591-FE22-11D0-8BEF-0060081841DE}")
        self._voiceNo = sapi4.Find(0)
        sapi4.Select(self._voiceNo)
        sapi4.Speak(" ")
        self.__object = sapi4
        self._voice_list = self._available_voices()

    def _set_capabilities(self):
        sapi4 = self.__object
        try:
            sapi4.Pitch = sapi4.Pitch
            self._min_pitch = sapi4.MinPitch
            self._max_pitch = sapi4.MaxPitch
            self._has_pitch = True
        except:
            self._min_pitch = 0
            self._max_pitch = 0
            self._has_pitch = False
        try:
            sapi4.Speed = sapi4.Speed
            self._min_rate = sapi4.MinSpeed
            self._max_rate = sapi4.MaxSpeed
            self._has_rate = True
        except:
            self._min_rate = 0
            self._max_rate = 0
            self._has_rate = False
        try:
            sapi4.VolumeLeft = sapi4.VolumeLeft
            self._min_volume = sapi4.MinVolumeLeft
            self._max_volume = sapi4.MaxVolumeLeft
            self._has_volume = True
        except:
            self._min_volume = 0
            self._max_volume = 0
            self._has_volume = False

    def _available_voices(self):
        voice_list = []
        for voice_no in range(1, self.__object.CountEngines):
            voice_list.append(self.__object.ModeName(voice_no))
        return voice_list

    @property
    def available_voices(self):
        return self._voice_list

    def list_voices(self):
        return self.available_voices

    def get_voice(self):
        return self.__object.ModeName(self._voice_no)

    def set_voice(self, value):
        self._voice_no = self.list_voices().index(value) + 1
        self.__object.Select(self._voice_no)
        self.silence()
        self.__object.Speak(" ")
        self._set_capabilities()

    def get_pitch(self):
        if self.has_pitch:
            return self.__object.Pitch

    def set_pitch(self, value):
        if self.has_pitch:
            self.__object.Pitch = value

    def get_rate(self):
        if self.has_rate:
            return self.__object.Speed

    def set_rate(self, value):
        if self.has_rate:
            self.__object.Speed = value

    def get_volume(self):
        if self.has_volume:
            return self.__object.VolumeLeft

    def set_volume(self, value):
        if self.has_volume:
            self.__object.VolumeLeft = value

    @property
    def has_pitch(self):
        return self._has_pitch

    @property
    def has_rate(self):
        return self._has_rate

    @property
    def has_volume(self):
        return self._has_volume

    @property
    def min_pitch(self):
        return self._min_pitch

    @property
    def max_pitch(self):
        return self._max_pitch

    @property
    def min_rate(self):
        return self._min_rate

    @property
    def max_rate(self):
        return self._max_rate

    @property
    def min_volume(self):
        return self._min_volume

    @property
    def max_volume(self):
        return self._max_volume

    def speak(self, text, interrupt=False):
        if interrupt:
            self.silence()
        self.__object.Speak(text)

    def silence(self):
        self.__object.AudioReset()


output_class = Sapi4



from __future__ import absolute_import
import os
import platform
import ctypes

from platform_utils import paths
from libloader import load_library
from .base import Output


class ZDSR(Output):
    """Supports The ZDSR screen reader"""

    name = "ZDSR"
    lib32 = "ZDSRAPI.dll"
    lib64 = "ZDSRAPI_x64.dll"
    is_loaded=False
    argtypes = {
        "speak": (ctypes.c_wchar_p,),
    }

    def load(self):
        self.lib.InitTTS(0,"")
        self.is_loaded=True

    def is_active(self):
        try:
            if not self.is_loaded: self.load()
            return self.lib.GetSpeakState() != 2
        except:
            return False

    def speak(self, text, interrupt=False):
        if not self.is_loaded: self.load()
        self.lib.Speak(text,interrupt)

    def silence(self):
        if not self.is_loaded: self.load()
        self.lib.StopSpeak()


output_class = ZDSR



from __future__ import absolute_import
import platform

if platform.system() == "Windows":
    from libloader import com
    from libloader.com import load_com

    def _load_com(*names):
        try:
            return load_com(*names)
        except AttributeError:
            # remove cache
            import os
            import sys
            import shutil
            for module in [m.__name__ for m in sys.modules.values()]:
                if module.startswith("win32com.gen_py."):
                    del sys.modules[module]
            shutil.rmtree(os.path.join(os.environ.get('LOCALAPPDATA'), 'Temp', 'gen_py'))
            # try again
            return load_com(*names)
    com.load_com = _load_com

    from . import nvda
    from . import jaws
    from . import sapi5
    from . import window_eyes
    from . import system_access
    from . import dolphin
    from . import pc_talker
    from . import zdsr

    # import sapi4

if platform.system() == "Darwin":
    from . import voiceover
    from . import nsspeechsynthesizer

if platform.system() == "Linux":
    from . import speech_dispatcher
    from . import e_speak

from . import auto



from __future__ import absolute_import
from collections import OrderedDict
from libloader.com import load_com
from .base import Output, OutputError
import pywintypes
import logging

log = logging.getLogger(__name__)

SVSFDefault = 0
SVSFlagsAsync = 1
SVSFPurgeBeforeSpeak = 2
SVSFIsFilename = 4
SVSFIsXML = 8
SVSFIsNotXML = 16
SVSFPersistXML = 32


class SAPI5(Output):
    """Supports the microsoft speech API version 5."""

    has_volume = True
    has_rate = True
    has_pitch = True
    min_pitch = -10
    max_pitch = 10
    min_rate = -10
    max_rate = 10
    min_volume = 0
    max_volume = 100
    name = "sapi5"
    priority = 101
    system_output = True

    def __init__(self):
        try:
            self.object = load_com("SAPI.SPVoice")
            self._voices = self._available_voices()
        except (pywintypes.com_error, TypeError):
            raise OutputError
        self._pitch = 0

    def _available_voices(self):
        _voices = OrderedDict()
        for v in self.object.GetVoices():
            _voices[v.GetDescription()] = v
        return _voices

    def list_voices(self):
        return list(self._voices.keys())

    def get_voice(self):
        return self.object.Voice.GetDescription()

    def set_voice(self, value):
        log.debug('Setting SAPI5 voice to "%s"' % value)
        self.object.Voice = self._voices[value]
        # For some reason SAPI5 does not reset audio after changing the voice
        # By setting the audio device after changing voices seems to fix this
        # This was noted from information at:
        # http://lists.nvaccess.org/pipermail/nvda-dev/2011-November/022464.html
        self.object.AudioOutput = self.object.AudioOutput

    def get_pitch(self):
        return self._pitch

    def set_pitch(self, value):
        log.debug("Setting pitch to %d" % value)
        self._pitch = value

    def get_rate(self):
        return self.object.Rate

    def set_rate(self, value):
        log.debug("Setting rate to %d" % value)
        self.object.Rate = value

    def get_volume(self):
        return self.object.Volume

    def set_volume(self, value):
        self.object.Volume = value

    def speak(self, text, interrupt=False):
        if interrupt:
            self.silence()
        # We need to do the pitch in XML here
        textOutput = '<pitch absmiddle="%d">%s</pitch>' % (
            round(self._pitch),
            text.replace("<", "&lt;"),
        )
        self.object.Speak(textOutput, SVSFlagsAsync | SVSFIsXML)

    def silence(self):
        self.object.Speak("", SVSFlagsAsync | SVSFPurgeBeforeSpeak)

    def is_active(self):
        if self.object:
            return True
        return False


output_class = SAPI5



from __future__ import absolute_import
import os
import platform
import ctypes

from platform_utils import paths
from libloader import load_library
from .base import Output


class NVDA(Output):
    """Supports The NVDA screen reader"""

    name = "NVDA"
    lib32 = "nvdaControllerClient32.dll"
    lib64 = "nvdaControllerClient64.dll"
    argtypes = {
        "nvdaController_brailleMessage": (ctypes.c_wchar_p,),
        "nvdaController_speakText": (ctypes.c_wchar_p,),
    }

    def is_active(self):
        try:
            return self.lib.nvdaController_testIfRunning() == 0
        except:
            return False

    def braille(self, text, **options):
        self.lib.nvdaController_brailleMessage(text)

    def speak(self, text, interrupt=False):
        if interrupt:
            self.silence()
        self.lib.nvdaController_speakText(text)

    def silence(self):
        self.lib.nvdaController_cancelSpeech()


output_class = NVDA



from accessible_output3 import load_library
import platform


class OutputError(Exception):
    """Raised by outputs if they fail to initialize or output"""
    pass


class Output(object):
    """The base Output object.
    Most functionality is found in child classes.
    If wishing to implement support for a new output, it might be helpful to take a look at an existing one to see how everything works.
    """

    name = "Unnamed Output"
    """str: Name of the output. This attribute should be defined in all child classes for usability purposes."""

    lib32 = None
    """str: Path pointing to a 32 bit DLL used to interface with the output. Can be None."""

    lib64 = None
    """str: Path pointing to a 64 bit DLL used to interface with the output. Can be None."""

    argtypes = {}
    """dict: A mapping of funtion names (str) to signatures (tuple) containing C types."""

    cdll = False
    """bool: Whether the DLL uses __cdecl calling conventions and should be loaded accordingly."""

    priority = 100
    """int: The priority this output should be given when automatically choosing possible outputs.
    See :class:`accessible_output3.outputs.auto.Auto`.
    Priority begins at 100. Anything higher will throw it lower on the list of available outputs. Anything lower will make it more likely to be the first option selected.
    Example, screen readers are sometimes given a higher priority (usuallly 100) than TTS such as SAAPI5 (101) due to a common usability preference."""

    system_output = False
    """bool: Is this output present and accessible on most systems with little or no additional installation or configuration?
    example: SAPI5 is a system output, while NVDA is not."""

    def __init__(self):
        self.is_32bit = platform.architecture()[0] == "32bit"
        if self.lib32 and self.is_32bit:
            self.lib = load_library(self.lib32, cdll=self.cdll)
        elif self.lib64:
            self.lib = load_library(self.lib64, cdll=self.cdll)
        else:
            self.lib = None
        if self.lib is not None:
            for func in self.argtypes:
                try:
                    getattr(self.lib, func).argtypes = self.argtypes[func]
                except AttributeError:
                    pass

    def output(self, text, **options):
        """
        Output the given text in speech, braille or both depending on what the output supports

        Args:
          text (str): The text to output.
          **options: Additional options.

        raises:
            RuntimeError: If the requested output doesn't define either speak or braille.
        """
        output = False
        if self.speak(text, **options):
            output = True
        if self.braille(text, **options):
            output = True
        if not output:
            raise RuntimeError(
                "Output %r does not have any method defined to output" % self
            )

    def is_system_output(self):
        return self.system_output

    def speak(self, text, **options):
        """
        Speaks the given text.

        Args:
          text (str): The text to speak.
          **options: Additional options.
        """
        return False

    def braille(self, text, **options):
        """
        Brailles the given text.

        Args:
          text (str): The text to braille.
          **options: Additional options.
        """
        return False



from __future__ import absolute_import
from .base import Output

class ESpeak(Output):
    """Supports ESpeak on Linux

    Note this requires python-espeak to be installed
    This can be done on Debian distros by using apt-get install python-espeak
    Or through `this tarball <https://launchpad.net/python-espeak>`_.
	"""

    name = "Linux ESpeak"
    priority = 101
    _ec = None

    def __init__(self):
        try:
            import espeak.core
            self._ec = espeak.core
        except:
            print("Cannot find espeak.core. Please install python-espeak or python3-espeak.")

    def is_active(self):
        return self._ec is not None

    def speak(self, text, interrupt=0):
        if interrupt:
            self.silence()
        self._ec.synth(text)

    def silence(self):
        self._ec.cancel()


output_class = ESpeak



from __future__ import absolute_import
import ctypes
from .base import Output


class SystemAccess(Output):
    """Supports System Access and System Access Mobile"""

    name = "System Access"
    lib32 = "saapi32.dll"
    argtypes = {"SA_BrlShowTextW": (ctypes.c_wchar_p,), "SA_SayW": (ctypes.c_wchar_p,)}
    priority = 99

    def braille(self, text, **options):
        self.lib.SA_BrlShowTextW(text)

    def speak(self, text, interrupt=False):
        if self.is_active():
            self.dll.SA_SayW(str(text))

    def is_active(self):
        try:
            return self.dll.SA_IsRunning()
        except:
            return False


output_class = SystemAccess



from __future__ import absolute_import
import os
import ctypes

from .base import Output


class Dolphin(Output):
    """Supports dolphin products."""

    name = "Dolphin"
    lib32 = "dolapi.dll"
    argtypes = {
        "DolAccess_Command": (ctypes.c_wchar_p, ctypes.c_int, ctypes.c_int),
        "DolAccess_Action": (ctypes.c_int,),
    }

    def speak(self, text, interrupt=0):
        if interrupt:
            self.silence()
        # If we don't call this, the API won't let us speak.
        if self.is_active():
            self.lib.DolAccess_Command(text, (len(text) * 2) + 2, 1)

    def silence(self):
        self.lib.DolAccess_Action(141)

    def is_active(self):
        try:
            return self.lib.DolAccess_GetSystem() in (1, 4, 8)
        except:
            return False


output_class = Dolphin



from __future__ import absolute_import

from .base import Output

class VoiceOver(Output):

    """Speech output supporting the Apple VoiceOver screen reader."""

    name = "VoiceOver"

    def __init__(self, *args, **kwargs):
        import appscript
        self.app = appscript.app("voiceover")

    def speak(self, text, interrupt=False):
        self.app.output(text)

    def silence(self):
        self.app.output(u"")

    def is_active(self):
        return self.app.isrunning()


output_class = VoiceOver



from setuptools import setup, find_packages
import io

__doc__ = (
    """Library to provide speech and braille output to a variety of different screen readers and other accessibility solutions."""
)

with io.open("readme.rst", encoding="UTF8") as readme:
    long_description = readme.read()

setup(
    name="accessible_output3",
    author="Luna R",
    author_email="pypi@lvna.me",
    version="0.1.1",
    description=__doc__,
    long_description=long_description,
    package_dir={"accessible_output3": "accessible_output3"},
    packages=find_packages(),
    package_data={"accessible_output3": ["lib/*"]},
    zip_safe=False,
    install_requires=["libloader", "platform_utils"],
    extras_require={
        ':sys_platform == "win32"': ["pywin32", "libloader"],
        ':sys_platform == "darwin"': ["appscript"],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Topic :: Adaptive Technologies",
        "Topic :: Software Development :: Libraries",
    ],
)



from __future__ import absolute_import
import accessible_output3
from .base import Output, OutputError


class Auto(Output):
    """An output which automatically selects the first available output on the system"""
    def __init__(self):
        output_classes = accessible_output3.get_output_classes()
        self.outputs = []
        for output in output_classes:
            try:
                self.outputs.append(output())
            except OutputError:
                pass

    def get_first_available_output(self):
        """
        Finds the ffirst available output.
        This is automatically called in braille, output and speak.
        """
        for output in self.outputs:
            if output.is_active():
                return output
        return None

    def speak(self, *args, **kwargs):
        """
        Speaks the given text if the output supports speech

        Args:
          text (str): The text to speak.
          **options: Additional options.
        """
        output = self.get_first_available_output()
        if output:
            output.speak(*args, **kwargs)

    def braille(self, *args, **kwargs):
        """
        Brailles the given text if the output supports Braille

        Args:
          text (str): The text to braille.
          **options: Additional options.
        """
        output = self.get_first_available_output()
        if output:
            output.braille(*args, **kwargs)

    def output(self, *args, **kwargs):
        output = self.get_first_available_output()
        if output:
            output.speak(*args, **kwargs)

    def is_system_output(self):
        """Returns True if this output is a system output."""
        output = self.get_first_available_output()
        if output:
            return output.is_system_output()
